/* ============================================================================== 
 * Copyright (c) 2013 Pengchong Jin (Purdue University)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice, this
 * list of conditions and the following disclaimer in the documentation and/or
 * other materials provided with the distribution.
 *
 * Neither the name of Pengchong Jin, Purdue University,
 * nor the names of its contributors may be used
 * to endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ============================================================================== */

#include <math.h>

#include "data2d.h"
#include "icd2d.h"

ENTRY ppotential(ENTRY delta, ENTRY p)
{
	ENTRY ppot;

	ppot = pow(fabs(delta), p)/p;

	return ppot;
}

ENTRY quadUpdate(struct ICDInfo *icd_info)
{
	int j;
	ENTRY num, denom;
	ENTRY sum1, sum2;

	denom = icd_info->th2 + 1.0/(icd_info->sigma*icd_info->sigma);
	sum1 = 0.0;
	sum2 = 0.0;
	for (j = 0; j < 4; j++)
	{
		sum1 += icd_info->neighbors[j];
	}
	for (; j < 8; j++)
	{
		sum2 += icd_info->neighbors[j];
	}

	num = (W1*sum1 + W2*sum2)/(icd_info->sigma*icd_info->sigma) + icd_info->th2*icd_info->v - icd_info->th1;

	return num/denom;
}

ENTRY qpotential(ENTRY delta, ENTRY p, ENTRY q, ENTRY c)
{
	ENTRY ppot, num;

	ppot = pow(fabs(delta), q)/q;
	num = pow(fabs(delta/c), p-q);

	return ppot*num/(1.0+num);
}

/* not rho'(delta), but rho'(delta)/delta */
ENTRY qinfluence(ENTRY delta, ENTRY p, ENTRY q, ENTRY c)
{
	ENTRY pmq;
	ENTRY fact, denom, sum1, sum2;

	pmq = p - q;
	fact = q*pow(c, pmq);

	if (delta == 0.0)
	{
		return 2.0/fact;
	}

	denom = 1.0 + pow(fabs(delta/c), pmq);
	sum1 = 2.0/denom;
	sum2 = delta*pmq*pow(fabs(delta), 1.0-q)/(pow(c, pmq)*denom*denom);

	if (delta < 0.0)
	{
		return (sum1 + sum2)/fact;
	}
	else
	{
		return (sum1 - sum2)/fact;
	}
}

ENTRY qGGMRFUpdate(struct ICDInfo *icd_info)
{
	int j;
	ENTRY num, denom, c, delta, influ;
	ENTRY num1, num2, denom1, denom2;

	num1 = 0.0;
	num2 = 0.0;
	denom1 = 0.0;
	denom2 = 0.0;
	for (j = 0; j < 8; j++)
	{
		delta = (icd_info->v - icd_info->neighbors[j])/icd_info->sigma;
		c = icd_info->c/icd_info->sigma;
		influ = qinfluence(delta, icd_info->p, icd_info->q, c);

		if (j < 4)
		{
			num1 += (influ*icd_info->neighbors[j]);
			denom1 += influ;
		}
		if (j >= 4)
		{
			num2 += (influ*icd_info->neighbors[j]);
			denom2 += influ;
		}
	}

	num = (W1*num1 + W2*num2)/(icd_info->sigma*icd_info->sigma) + icd_info->th2*icd_info->v - icd_info->th1;
	denom = (W1*denom1 + W2*denom2)/(icd_info->sigma*icd_info->sigma) + icd_info->th2;

	return num/denom;
}

ENTRY ICDStep2D(
	ENTRY *y,
	ENTRY *Ax,
	struct ICDInfo *icd_info,
	struct SinoInfo *sino_info,
	struct ACol *A_col)
{
	int r, i;
	ENTRY w;
	ENTRY opt;

	/* compute the first and second derivatives */
	icd_info->th1 = 0.0;
	icd_info->th2 = 0.0;
	for (r = 0; r < A_col->n_index; r++)
	{
		i = A_col->index[r];

		if (icd_info->weight == 1)
		{
			w = exp(-y[i]/2.0)*sino_info->norm_factor;
			icd_info->th1 += A_col->val[r]*w*(Ax[i]-y[i]);
			icd_info->th2 += A_col->val[r]*w*A_col->val[r];
		}
		else
		{
			icd_info->th1 += A_col->val[r]*(Ax[i]-y[i]);
			icd_info->th2 += A_col->val[r]*A_col->val[r];
		}
	}

	/* compute the ML estimate */
	icd_info->ml = icd_info->v - icd_info->th1/icd_info->th2;
	if (icd_info->est == 0)
	{
		return icd_info->ml;
	}

	if (icd_info->est == 1)
	{
		opt = quadUpdate(icd_info);
	}
	else if (icd_info->est == 2)
	{
		opt = qGGMRFUpdate(icd_info);
	}

	return opt;
}
