/* ============================================================================== 
 * Copyright (c) 2013 Pengchong Jin (Purdue University)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice, this
 * list of conditions and the following disclaimer in the documentation and/or
 * other materials provided with the distribution.
 *
 * Neither the name of Pengchong Jin, Purdue University,
 * nor the names of its contributors may be used
 * to endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ============================================================================== */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>	/* strcmp */
#include <getopt.h>	/* getopt */

#include "data2d.h"
#include "allocate.h"
#include "io2d.h"

void readImgInfo(char *fname, struct ImgInfo *img_info)
{
	FILE *fp;
	char tag[200];

	if ((fp = fopen(fname, "r")) == NULL)
	{
		fprintf(stderr, "ERROR in readImgInfo: can't open file %s.\n", fname);
		exit(-1);
	}

	fgets(tag, 200, fp);
	fscanf(fp, "%d\n\n", &(img_info->Nx));
	fgets(tag, 200, fp);
	fscanf(fp, "%d\n\n", &(img_info->Ny));
	fgets(tag, 200, fp);
	fscanf(fp, "%lf\n\n", &(img_info->fov));
	fgets(tag, 200, fp);
	fscanf(fp, "%lf\n\n", &(img_info->r));
	fgets(tag, 200, fp);
	fscanf(fp, "%lf\n\n", &(img_info->mu_water));
	fgets(tag, 200, fp);
	fscanf(fp, "%d\n\n", &(img_info->upsample));
	fgets(tag, 200, fp);
	fscanf(fp, "%s\n", img_info->imgFile);

	fclose(fp);
}

void printImgInfo(struct ImgInfo *img_info)
{
	fprintf(stdout, "\nIMAGE PARAMETERS:\n");
	fprintf(stdout, " - Nx       = %d\n", img_info->Nx);
	fprintf(stdout, " - Ny       = %d\n", img_info->Ny);
	fprintf(stdout, " - fov      = %f (mm)\n", img_info->fov);
	fprintf(stdout, " - r        = %f (mm)\n", img_info->r);
	fprintf(stdout, " - delta    = %f (mm)\n", img_info->delta);
	fprintf(stdout, " - x_0      = %f (mm)\n", img_info->x_0);
	fprintf(stdout, " - y_0      = %f (mm)\n", img_info->y_0);
	fprintf(stdout, " - mu_water = %f (mm-1)\n", img_info->mu_water);
	fprintf(stdout, " - upsample = %d (0 = const, 1 = no upsample, x = upsample factor)\n", img_info->upsample);
	fprintf(stdout, " - imgFile  = %s\n\n", img_info->imgFile);
}

void constImage(ENTRY **img, ENTRY mu, struct ImgInfo *img_info)
{
	int j, Nx, Ny, N;

	Nx = img_info->Nx;
	Ny = img_info->Ny;
	N = Nx*Ny;

	*img = (ENTRY *)get_spc(N, sizeof(ENTRY));
	for (j = 0; j < N; j++)
	{
		(*img)[j] = mu;
	}
}

void upsampleImage(ENTRY **img, ENTRY *simg, int fact, struct ImgInfo *img_info)
{
	int jx, jy, rx, ry, Nx, Ny, N;
	ENTRY pixel;

	Nx = img_info->Nx;
	Ny = img_info->Ny;
	N = Nx*Ny;

	*img = (ENTRY *)get_spc(N, sizeof(ENTRY));
	for (jy = 0; jy < Ny/fact; jy++)
	{
		for (jx = 0; jx < Nx/fact; jx++)
		{
			pixel = simg[jy*Nx/fact+jx];
			for (ry = 0; ry < fact; ry++)
			{
				for (rx = 0; rx < fact; rx++)
				{
					(*img)[(jy*fact+ry)*Nx+jx*fact+rx] = pixel;
				}
			}
		}
	}
}

void readImage(char *fname, ENTRY **img, int N, ENTRY mu_water)
{
	FILE *fp;
	int j;
	ENTRY pixel;

	if ((fp = fopen(fname, "r")) == NULL)
	{
		fprintf(stderr, "ERROR in readImage: can't open file %s.\n", fname);
		exit(-1);
	}

	*img = (ENTRY *)get_spc(N, sizeof(ENTRY));
	for (j = 0; j < N; j++)
	{
		fscanf(fp, "%lf ", &pixel);
		(*img)[j] = hu2mu(pixel, ATT_AIR, mu_water);
	}
	fclose(fp);
}

void maskImage(struct SinoInfo *sino_info, struct Image *image)
{
	int jx, jy, Nx, Ny;
	ENTRY x_0, y_0, delta, x, y, yy, r_sq, r_sq_max;

	Nx = image->img_info.Nx;
	Ny = image->img_info.Ny;
	x_0 = image->img_info.x_0;
	y_0 = image->img_info.y_0;
	delta = image->img_info.delta;

	image->recon_mask = (char **)multialloc(sizeof(char), 2, Ny, Nx);

	if (image->img_info.r < 0.0)
	{
		r_sq_max = sino_info->t_0*sino_info->t_0;
	}
	else
	{
		r_sq_max = image->img_info.r*image->img_info.r;
	}

	for (jy = 0; jy < Ny; jy++)
	{
		y = y_0 + jy*delta;
		yy = y*y;
		for (jx = 0; jx < Nx; jx++)
		{
			x = x_0 + jx*delta;

			r_sq = x*x + yy;
			if (r_sq > r_sq_max)
			{
				image->recon_mask[jy][jx] = 0;
				image->img[jy*Nx+jx] = ATT_AIR;
			}
			else
			{
				image->recon_mask[jy][jx] = 1;
			}
		}
	}
}

void initializeImage(struct SinoInfo *sino_info, struct Image *image)
{
	ENTRY *simg;

	if (image->img_info.upsample == 0)
	{
		constImage(&(image->img), image->img_info.mu_water, &(image->img_info));
	}
	else if (image->img_info.upsample == 1)
	{
		readImage(image->img_info.imgFile, &(image->img), image->img_info.Nx*image->img_info.Ny, image->img_info.mu_water);
	}
	else if (image->img_info.upsample > 1)
	{
		readImage(image->img_info.imgFile, &simg, image->img_info.Nx*image->img_info.Ny/(image->img_info.upsample*image->img_info.upsample), image->img_info.mu_water);
		upsampleImage(&(image->img), simg, image->img_info.upsample, &(image->img_info));
		free((void *)simg);
	}

	maskImage(sino_info, image);
}

void writeImage(char *fname, struct Image *image)
{
	FILE *fp;
	int j, N;
	ENTRY pixel;
	ENTRY mu_water;

	if ((fp = fopen(fname, "w")) == NULL)
	{
		fprintf(stderr, "ERROR in writeImage: can't open file %s.\n\n", fname);
		exit(-1);
	}

	mu_water = image->img_info.mu_water;
	N = image->img_info.Nx*image->img_info.Ny;
	for (j = 0; j < N; j++)
	{
		pixel = mu2hu(image->img[j], ATT_AIR, mu_water);
		if (pixel > 0.0)
		{
			fprintf(fp, "%g ", pixel);
		}
		else
		{
			fprintf(fp, "0 ");
		}
		if ((j+1) % image->img_info.Nx == 0)
		{
			fprintf(fp, "\n");
		}
	}
	fclose(fp);
}

void freeImage(struct Image *image)
{
	free((void *)image->img);
	multifree((void *)image->recon_mask, 2);
}

void readSinoInfo(char *fname, struct SinoInfo *sino_info)
{
	FILE *fp;
	char tag[200];

	if ((fp = fopen(fname, "r")) == NULL)
	{
		fprintf(stderr, "ERROR in readSinoInfo: can't open file %s.\n", fname);
		exit(-1);
	}

	fgets(tag, 200, fp);
	fscanf(fp, "%d\n\n", &(sino_info->N_theta));
	fgets(tag, 200, fp);
	fscanf(fp, "%d\n\n", &(sino_info->N_t));
	fgets(tag, 200, fp);
	fscanf(fp, "%lf\n\n", &(sino_info->delta_theta));
	fgets(tag, 200, fp);
	fscanf(fp, "%lf\n\n", &(sino_info->delta_t));
	fgets(tag, 200, fp);
	fscanf(fp, "%lf\n\n", &(sino_info->theta_0));
	fgets(tag, 200, fp);
	fscanf(fp, "%lf\n\n", &(sino_info->t_0));
	fgets(tag, 200, fp);
	fscanf(fp, "%s\n", sino_info->sinoFile);

	fclose(fp);
}

void printSinoInfo(struct SinoInfo *sino_info)
{
	fprintf(stdout, "\nSINOGRAM PARAMETERS:\n");
	fprintf(stdout, " - N_theta     = %d\n", sino_info->N_theta);
	fprintf(stdout, " - N_t         = %d\n", sino_info->N_t);
	fprintf(stdout, " - delta_theta = %f (rad)\n", sino_info->delta_theta);
	fprintf(stdout, " - delta_t     = %f (mm)\n", sino_info->delta_t);
	fprintf(stdout, " - theta_0     = %f (rad)\n", sino_info->theta_0);
	fprintf(stdout, " - t_0         = %f (mm)\n", sino_info->t_0);
	fprintf(stdout, " - sinoFile    = %s\n\n", sino_info->sinoFile);
}

void readSinogram(char *fname, struct Sinogram *sinogram)
{
	FILE *fp;
	int i, M;
	struct SinoInfo *sino_info;
	ENTRY *sino;

	sino_info = &(sinogram->sino_info);

	if ((fp = fopen(fname, "r")) == NULL)
	{
		fprintf(stderr, "ERROR in readSingram: can't open file %s.\n", fname);
		exit(-1);
	}

	M = sino_info->N_theta*sino_info->N_t;
	sinogram->sino = (ENTRY *)get_spc(M, sizeof(ENTRY));
	sino = sinogram->sino;
	for (i = 0; i < M; i++)
	{
		fscanf(fp, "%lf ", &(sino[i]));
	}
	fclose(fp);
}

void writeSinogram(char *fname, struct Sinogram *sinogram)
{
	FILE *fp;
	int i, M;
	ENTRY pixel;

	if ((fp = fopen(fname, "w")) == NULL)
	{
		fprintf(stderr, "ERROR in writeSinogram: can't open file %s.\n", fname);
		exit(-1);
	}

	M = sinogram->sino_info.N_theta*sinogram->sino_info.N_t;
	for (i = 0; i < M; i++)
	{
		pixel = sinogram->sino[i];
		if (pixel > 0.0)
		{
			fprintf(fp, "%g ", pixel);
		}
		else
		{
			fprintf(fp, "0 ");
		}
		if ((i+1) % sinogram->sino_info.N_t == 0)
		{
			fprintf(fp, "\n");
		}
	}
	fclose(fp);
}

void freeSinogram(struct Sinogram *sinogram)
{
	free((void *)sinogram->sino);
}

void readPriorInfo(char *fname, struct PriorInfo *prior_info)
{
	FILE *fp;
	char tag[200];

	if ((fp = fopen(fname, "r")) == NULL)
	{
		fprintf(stderr, "ERROR in readPriorInfo: can't open file %s.\n", fname);
		exit(-1);
	}

	fgets(tag, 200, fp);
	fscanf(fp, "%lf\n\n", &(prior_info->p));
	fgets(tag, 200, fp);
	fscanf(fp, "%lf\n\n", &(prior_info->q));
	fgets(tag, 200, fp);
	fscanf(fp, "%lf\n\n", &(prior_info->c));
	fgets(tag, 200, fp);
	fscanf(fp, "%lf\n\n", &(prior_info->sigma));
	fgets(tag, 200, fp);
	fscanf(fp, "%d\n\n", &(prior_info->weight));
	fgets(tag, 200, fp);
	fscanf(fp, "%lf\n\n", &(prior_info->stop_threshold));
	fgets(tag, 200, fp);
	fscanf(fp, "%d\n\n", &(prior_info->Nit));
	fgets(tag, 200, fp);
	fscanf(fp, "%s\n", prior_info->oname);

	fclose(fp);
}

void printPriorInfo(struct PriorInfo *prior_info, ENTRY mu_water)
{
	fprintf(stdout, "\nPRIOR PARAMETERS:\n");
	fprintf(stdout, " - est            = %d (0 = ML, 1 = quadratic, 2 = q-GGMRF)\n", prior_info->est);
	fprintf(stdout, " - p              = %f\n", prior_info->p);
	fprintf(stdout, " - q              = %f\n", prior_info->q);
	fprintf(stdout, " - c              = %f (mm-1) (%f HU)\n", prior_info->c, mu2hu(prior_info->c, ATT_AIR, mu_water));
	fprintf(stdout, " - sigma          = %f\n", prior_info->sigma);
	fprintf(stdout, " - weight         = %d (0 = identity, 1 = exponential)\n", prior_info->weight);
	fprintf(stdout, " - stop_threshold = %f (mm-1) (%f HU) (0 = no stopping condition)\n", prior_info->stop_threshold, mu2hu(prior_info->stop_threshold, ATT_AIR, mu_water));
	fprintf(stdout, " - Nit            = %d\n", prior_info->Nit);
	fprintf(stdout, " - oname          = %s\n\n", prior_info->oname);
}

void usage(char *fname)
{
	fprintf(stdout, "\n MBIR RECONSTRUCTION SOFTWARE FOR 2D PARALLEL-BEAM X-RAY CT\n");
	fprintf(stdout, " writen by Pengchong Jin, Purdue University\n");
	fprintf(stdout, " build time: %s, %s\n", __DATE__,  __TIME__);
	fprintf(stdout, " usage:\n");
	fprintf(stdout, " - (a): %s -i <img param> -j <sino param> -k <prior param>\n", fname);
	fprintf(stdout, "           see sample files for the format.\n");
	fprintf(stdout, " - (b): %s -S <sino file> -A <N_theta> -T <N_t> -a <delta_theta> -t <delta_t> -O <ouput file> [optional args]\n", fname);
	fprintf(stdout, "           -x: number of pixels in x, (N_t)\n");
	fprintf(stdout, "           -y: number of pixels in y, should be equal to number of pixels in x, (N_t)\n");
	fprintf(stdout, "           -f: field of view in x (and y) (mm), (N_t*delta_t)\n");
	fprintf(stdout, "           -r: radius of the reconstruction (mm), (|t_0|)\n");
	fprintf(stdout, "           -w: linear attenuation of water (mm-1), (0.0202527 mm-1)\n");
	fprintf(stdout, "           -u: upsample options for initial image, 0 = const, 1 = no upsample, x = factor x, (0)\n");
	fprintf(stdout, "           -I: initial image file name, NA = no initial, use const, (NA)\n");
	fprintf(stdout, "           -b: initial projection angle (rad), (0)\n");
	fprintf(stdout, "           -s: start of the displacement (mm), (-(N_t-1)*delta_t/2)\n");
	fprintf(stdout, "           -p: q-GGMRF p, either -1 (ML) or 2 (regularized q-GGMRF), (2)\n");
	fprintf(stdout, "           -q: q-GGMRF q, between 1.0 and 2.0, (1.2)\n");
	fprintf(stdout, "           -c: q-GGMRF c, (15 HU)\n");
	fprintf(stdout, "           -X: regularization factor, (0.8)\n");
	fprintf(stdout, "           -W: noise model weighting, (0 = idenity, 1 = exponential) (1)\n");
	fprintf(stdout, "           -h: stopping threshold, 0 = no stopping condition, (1 HU)\n");
	fprintf(stdout, "           -N: number of iterations, (20)\n\n");

	exit(-1);
}

void readData(
	int argc,
	char *argv[],
	struct Image *image,
	struct Sinogram *sinogram,
	struct PriorInfo *prior_info)
{
	char ch;
	char imgpFile[200];
	char sinopFile[200];
	char priorpFile[200];
	struct ImgInfo *img_info;
	struct SinoInfo *sino_info;

	img_info = &(image->img_info);
	sino_info = &(sinogram->sino_info);

	if (argc < 7)
	{
		usage(argv[0]);
	}

	img_info->Nx = -1;
	img_info->Ny = -1;
	img_info->fov = -1;
	img_info->mu_water = -1;
	img_info->r = -1;
	img_info->upsample = 0;
	sprintf(img_info->imgFile, "NA");
	sino_info->theta_0 = 0;
	sino_info->t_0 = -1;
	prior_info->est = 2;
	prior_info->p = 2;
	prior_info->q = 1.2;
	prior_info->c = -1;
	prior_info->sigma = 0.8;
	prior_info->weight = 1;
	prior_info->stop_threshold = -1;
	prior_info->Nit = 20;
	sprintf(imgpFile, "NA");
	sprintf(sinopFile, "NA");
	sprintf(priorpFile, "NA");

	/* get options */
	while ((ch = getopt(argc, argv, "i:j:k:S:A:T:a:t:O:x:y:f:r:w:u:I:b:s:p:q:c:X:W:h:N:")) != EOF)
	{
		switch (ch)
		{
			case 'i':
			{
				sprintf(imgpFile, "%s", optarg);
				break;
			}
			case 'j':
			{
				sprintf(sinopFile, "%s", optarg);
				break;
			}
			case 'k':
			{
				sprintf(priorpFile, "%s", optarg);
				break;
			}
			case 'S':	/* sinoFile */
			{
				sprintf(sino_info->sinoFile, "%s", optarg);
				break;
			}
			case 'A':	/* N_theta */
			{
				sscanf(optarg, "%d", &(sino_info->N_theta));
				break;
			}
			case 'T':	/* N_t */
			{
				sscanf(optarg, "%d", &(sino_info->N_t));
				break;
			}
			case 'a':	/* delta_theta */
			{
				sscanf(optarg, "%lf", &(sino_info->delta_theta));
				break;
			}
			case 't':	/* delta_t */
			{
				sscanf(optarg, "%lf", &(sino_info->delta_t));
				break;
			}
			case 'O':	/* ouput file */
			{
				sprintf(prior_info->oname, "%s", optarg);
				break;
			}
			case 'x':	/* Nx */
			{
				sscanf(optarg, "%d", &(img_info->Nx));
				break;
			}
			case 'y':	/* Ny */
			{
				sscanf(optarg, "%d", &(img_info->Ny));
				break;
			}
			case 'f':	/* fov */
			{
				sscanf(optarg, "%lf", &(img_info->fov));
				break;
			}
			case 'r':	/* r */
			{
				sscanf(optarg, "%lf", &(img_info->r));
				break;
			}
			case 'w':	/* mu_water */
			{
				sscanf(optarg, "%lf", &(img_info->mu_water));
				break;
			}
			case 'u':	/* upsample */
			{
				sscanf(optarg, "%d", &(img_info->upsample));
				break;
			}
			case 'I':	/* imgFile */
			{
				sprintf(img_info->imgFile, "%s", optarg);
				break;
			}
			case 'b':	/* theta_0 */
			{
				sscanf(optarg, "%lf", &(sino_info->theta_0));
				break;
			}
			case 's':	/* t_0 */
			{
				sscanf(optarg, "%lf", &(sino_info->t_0));
				break;
			}
			case 'p':	/* q-GGMRF p */
			{
				sscanf(optarg, "%lf", &(prior_info->p));
				break;
			}
			case 'q':	/* q-GGMRF q */
			{
				sscanf(optarg, "%lf", &(prior_info->q));
				break;
			}
			case 'c':	/* q-GGMRF c */
			{
				sscanf(optarg, "%lf", &(prior_info->c));
				break;
			}
			case 'X':	/* regularization factor sigma */
			{
				sscanf(optarg, "%lf", &(prior_info->sigma));
				break;
			}
			case 'W':	/* noise model weighting, 0 = identity, 1 = exponential */
			{
				sscanf(optarg, "%d", &(prior_info->weight));
				break;
			}
			case 'h':	/* stop threshold */
			{
				sscanf(optarg, "%lf", &(prior_info->stop_threshold));
				break;
			}
			case 'N':	/* Nit */
			{
				sscanf(optarg, "%d", &(prior_info->Nit));
				break;
			}
			default:
			{
				usage(argv[0]);
				break;
			}
		}
	}

	if (strcmp(sinopFile, "NA") != 0)
	{
		readSinoInfo(sinopFile, &(sinogram->sino_info));
	}
	if (strcmp(imgpFile, "NA") != 0)
	{
		readImgInfo(imgpFile, &(image->img_info));
	}
	if (strcmp(priorpFile, "NA") != 0)
	{
		readPriorInfo(priorpFile, prior_info);
	}

	if (sino_info->t_0 == -1)
	{
		sino_info->t_0 = -(sino_info->N_t-1)*sino_info->delta_t/2.0;
	}
	if (img_info->Nx == -1)
	{
		img_info->Nx = sino_info->N_t;
	}
	if (img_info->Ny == -1)
	{
		img_info->Ny = sino_info->N_t;
	}
	if (img_info->Nx != img_info->Ny)
	{
		fprintf(stderr, "ERROR in readData: Nx does not equal Ny.\n");
		exit(-1);
	}
	if (img_info->fov == -1)
	{
		img_info->fov = sino_info->N_t*sino_info->delta_t;
	}
	if (img_info->mu_water == -1)
	{
		img_info->mu_water = ATT_H2O;
	}
	if (img_info->r == -1)
	{
		img_info->r = fabs(sino_info->t_0);
	}
	img_info->delta = img_info->fov/(ENTRY)img_info->Nx;
	img_info->x_0 = -(img_info->Nx-1)*img_info->delta/2.0;
	img_info->y_0 = -(img_info->Ny-1)*img_info->delta/2.0;
	if (prior_info->p == -1.0)
	{
		prior_info->est = 0;
	}
	else
	{
		if (prior_info->q == 2.0)
		{
			prior_info->est = 1;
		}
	}
	if (prior_info->p != -1.0 && prior_info->p != 2.0)
	{
		fprintf(stderr, "ERROR in readData: p should be either -1 (for ML estmation) or 2 (for regularizaed q-GGMRF prior).\n");
		exit(-1);
	}
	if (prior_info->q < 1.0 || prior_info->q > 2.0)
	{
		fprintf(stderr, "ERROR in readData: q should be in (1.0 2.0).\n");
		exit(-1);
	}

	if (prior_info->c < 0)
	{
		prior_info->c = 15.0;		/* default 15HU */
	}
	if (prior_info->stop_threshold < 0)
	{
		prior_info->stop_threshold = 1.0;
	}
	/* convert to mm-1 */
	prior_info->c = hu2mu(prior_info->c, ATT_AIR, img_info->mu_water);
	prior_info->stop_threshold = hu2mu(prior_info->stop_threshold, ATT_AIR, img_info->mu_water);

	/* read data */
	printSinoInfo(sino_info);
	readSinogram(sino_info->sinoFile, sinogram);

	printImgInfo(img_info);
	initializeImage(sino_info, image);

	printPriorInfo(prior_info, img_info->mu_water);
}

void freeData(struct Image *image, struct Sinogram *sinogram)
{
	freeImage(image);
	freeSinogram(sinogram);
}
