/* ============================================================================== 
 * Copyright (c) 2013 Pengchong Jin (Purdue University)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice, this
 * list of conditions and the following disclaimer in the documentation and/or
 * other materials provided with the distribution.
 *
 * Neither the name of Pengchong Jin, Purdue University,
 * nor the names of its contributors may be used
 * to endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ============================================================================== */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>

#include "data2d.h"
#include "allocate.h"
#include "A_comp.h"
#include "io2d.h"
#include "fpro2d.h"
#include "icd2d.h"
#include "recon2d.h"

void shuffle(int *order, int len)
{
	int i, j, tmp;

	srand(time(NULL));

	for (i = 0; i < len-1; i++)
	{
		j = i + (rand() % (len-i));
		tmp = order[j];
		order[j] = order[i];
		order[i] = tmp;
	}
}

void extractNeighbors2D(
	ENTRY *neighbors,
	int jx,
	int jy,
	ENTRY *x,
	struct ImgInfo *img_info)
{
	int plusx, minusx, plusy, minusy;
	int Nx, Ny;

	Nx = img_info->Nx;
	Ny = img_info->Ny;

	plusx = jx + 1;
	plusx = ((plusx < Nx) ? plusx : 0);
	minusx = jx - 1;
	minusx = ((minusx < 0) ? (Nx-1) : minusx);
	plusy = jy + 1;
	plusy = ((plusy < Ny) ? plusy : 0);
	minusy = jy - 1;
	minusy = ((minusy < 0) ? (Ny-1) : minusy);

	neighbors[0] = x[jy*Nx+plusx];
	neighbors[1] = x[jy*Nx+minusx];
	neighbors[2] = x[plusy*Nx+jx];
	neighbors[3] = x[minusy*Nx+jx];

	neighbors[4] = x[plusy*Nx+plusx];
	neighbors[5] = x[plusy*Nx+minusx];
	neighbors[6] = x[minusy*Nx+plusx];
	neighbors[7] = x[minusy*Nx+minusx];
}

ENTRY negLogPosterior2D(
	ENTRY *x,
	ENTRY *y,
	ENTRY *Ax,
	struct ImgInfo *img_info,
	struct SinoInfo *sino_info,
	struct PriorInfo *prior_info)
{
	int i, M, j, jx, jy, Nx, Ny, plusx, minusx, plusy;
	ENTRY w;
	ENTRY nloglike, nlogprior1, nlogprior2;

	M = sino_info->N_theta*sino_info->N_t;
	Nx = img_info->Nx;
	Ny = img_info->Ny;

	nloglike = 0.0;
	for (i = 0; i < M; i++)
	{
		if (prior_info->weight == 1)
		{
			w = exp(-y[i]/2.0)*sino_info->norm_factor;
			nloglike += (y[i]-Ax[i])*w*(y[i]-Ax[i]);
		}
		else
		{
			nloglike += (y[i]-Ax[i])*(y[i]-Ax[i]);
		}
	}
	nloglike /= 2.0;

	if (prior_info->est == 0)
	{
		return nloglike;
	}

	nlogprior1 = 0.0;
	nlogprior2 = 0.0;
	for (jy = 0; jy < Ny; jy++)
	{
		for (jx = 0; jx < Nx; jx++)
		{
			plusx = jx + 1;
			plusx = ((plusx < Nx) ? plusx : 0);
			minusx = jx - 1;
			minusx = ((minusx < 0) ? Nx-1 : minusx);
			plusy = jy + 1;
			plusy = ((plusy < Ny) ? plusy : 0);

			j = jy*Nx + jx;

			if (prior_info->est == 1)
			{
				nlogprior1 += ppotential((x[j] - x[jy*Nx+plusx])/prior_info->sigma, 2);
				nlogprior1 += ppotential((x[j] - x[plusy*Nx+jx])/prior_info->sigma, 2);

				nlogprior2 += ppotential((x[j] - x[plusy*Nx+minusx])/prior_info->sigma, 2);
				nlogprior2 += ppotential((x[j] - x[plusy*Nx+plusx])/prior_info->sigma, 2);
			}
			else if (prior_info->est == 2)
			{
				nlogprior1 += qpotential((x[j] - x[jy*Nx+plusx])/prior_info->sigma, prior_info->p, prior_info->q, prior_info->c/prior_info->sigma);
				nlogprior1 += qpotential((x[j] - x[plusy*Nx+jx])/prior_info->sigma, prior_info->p, prior_info->q, prior_info->c/prior_info->sigma);

				nlogprior2 += qpotential((x[j] - x[plusy*Nx+minusx])/prior_info->sigma, prior_info->p, prior_info->q, prior_info->c/prior_info->sigma);
				nlogprior2 += qpotential((x[j] - x[plusy*Nx+plusx])/prior_info->sigma, prior_info->p, prior_info->q, prior_info->c/prior_info->sigma);
			}
		}
	}

	return (nloglike + W1*nlogprior1 + W2*nlogprior2);
}

void updateAx2D(
	ENTRY *Ax,
	struct ACol *A_col,
	ENTRY diff)
{
	int r, i;

	for (r = 0; r < A_col->n_index; r++)
	{
		i = A_col->index[r];
		Ax[i] += A_col->val[r]*diff;
	}
}

void writeErrorSinogram(
	char *fname,
	ENTRY *y,
	ENTRY *Ax,
	struct SinoInfo *sino_info)
{
	FILE *fp;
	int i, M;

	M = sino_info->N_theta*sino_info->N_t;

	if ((fp = fopen(fname, "w")) == NULL)
	{
		fprintf(stderr, "ERROR in writeErrorSinogram: can't open file %s.\n", fname);
		exit(-1);
	}

	for (i = 0; i < M; i++)
	{
		fprintf(fp, "%f ", (y[i] - Ax[i]));

		if ((i+1) % sino_info->N_t == 0)
		{
			fprintf(fp, "\n");
		}
	}

	fclose(fp);
}

void writeNormalizedErrorSinogram(
	char *fname,
	ENTRY *y,
	ENTRY *Ax,
	struct SinoInfo *sino_info)
{
	FILE *fp;
	int i, M;

	M = sino_info->N_theta*sino_info->N_t;

	if ((fp = fopen(fname, "w")) == NULL)
	{
		fprintf(stderr, "ERROR in writeNormalizedErrorSinogram: can't open file %s.\n", fname);
		exit(-1);
	}

	for (i = 0; i < M; i++)
	{
		fprintf(fp, "%f ", (y[i] - Ax[i])*sqrt(exp(-y[i]/2.0)*sino_info->norm_factor));

		if ((i+1) % sino_info->N_t == 0)
		{
			fprintf(fp, "\n");
		}
	}

	fclose(fp);
}



void writeCost(
	char *fname,
	int it,
	ENTRY cost,
	ENTRY avg_update,
	ENTRY mu_water)
{
	FILE *fp;

	if ((fp = fopen(fname, "a")) == NULL)
	{
		fprintf(stderr, "ERROR in writeCost: can't open file %s.\n", fname);
		exit(-1);
	}

	fprintf(fp, "%d %.15f %.15f\n", it+1, cost, mu2hu(avg_update, ATT_AIR, mu_water));

	fclose(fp);
}

void MBIRReconstruct2D(
	struct Image *image,
	struct Sinogram *sinogram,
	struct PriorInfo *prior_info)
{
	int it, Nit, j, jj, jx, jy, Nx, Ny, N, i, M;
	ENTRY *x;
	ENTRY *y;
	ENTRY *Ax;
	struct ACol A_col;
	struct ImgInfo *img_info;
	struct SinoInfo *sino_info;
	struct ICDInfo icd_info;
	char **recon_mask;
	int *order;
	ENTRY count_total;
	ENTRY **profiles;
	ENTRY pixel, diff;
	ENTRY cost, total_updates, avg_update, stop_threshold;
	char zero_skip_FLAG;
	char stop_FLAG;
	char fname[200];

	img_info = &(image->img_info);
	sino_info = &(sinogram->sino_info);

	Nx = img_info->Nx;
	Ny = img_info->Ny;
	N = Nx*Ny;
	M = sino_info->N_theta*sino_info->N_t;
	recon_mask = image->recon_mask;

	x = image->img;
	y = sinogram->sino;
	Ax = (ENTRY *)get_spc(M, sizeof(ENTRY));
	A_col.val = (ENTRY *)get_spc(M, sizeof(ENTRY));
	A_col.index = (int *)get_spc(M, sizeof(int));

	order = (int *)get_spc(N, sizeof(int));
	for (j = 0; j < N; j++)
	{
		order[j] = j;
	}

	pix_prof_comp(sino_info, img_info, &profiles);

	forwardProject2D(Ax, x, profiles, &A_col, sino_info, img_info);

	count_total = 0.0;
	for (i = 0; i < M; i++)
	{
		count_total += exp(-y[i]);
	}
	sino_info->norm_factor = (ENTRY)M/count_total;

	icd_info.est = prior_info->est;
	icd_info.p = prior_info->p;
	icd_info.q = prior_info->q;
	icd_info.c = prior_info->c;
	icd_info.sigma = prior_info->sigma;
	icd_info.weight = prior_info->weight;
	Nit = prior_info->Nit;
	stop_threshold = prior_info->stop_threshold;

	stop_FLAG = 0;
	for (it = 0; ((it < Nit) && (stop_FLAG == 0)); it++)
	{
		fprintf(stdout, "it = %d\n", it+1);

		shuffle(order, N);

		total_updates = 0.0;
		for (jj = 0; jj < N; jj++)
		{
			jy = order[jj] / Nx;
			jx = order[jj] % Nx;
			if (recon_mask[jy][jx])
			{
				/* check if zero-skip */
				icd_info.v = (ENTRY)x[jy*Nx+jx];

				extractNeighbors2D(icd_info.neighbors, jx, jy, x, img_info);

				zero_skip_FLAG = 0;
				if (icd_info.v == 0.0)
				{
					zero_skip_FLAG = 1;	/* maybe skip */
					for (j = 0; j < 8; j++)
					{
						if (icd_info.neighbors[j] > 0.0)
						{
							zero_skip_FLAG = 0;
							break;
						}
					}
				}

				if (zero_skip_FLAG == 0)
				{
					/* compute A column */
					A_comp(jy, jx, sino_info, img_info, profiles, &A_col);

					if (A_col.n_index > 0)
					{
						pixel = ICDStep2D(y, Ax, &icd_info, sino_info, &A_col);

						x[jy*Nx+jx] = ((pixel < 0.0) ? 0.0 : pixel);

						diff = x[jy*Nx+jx] - icd_info.v;
						total_updates += fabs(diff);
						updateAx2D(Ax, &A_col, diff);
					}
				}
			}
		}

		/* Intermediate results */
/*		sprintf(fname, "%s_%d.rec", prior_info->oname, it+1);
		writeImage(fname, image);

		sprintf(fname, "%s_%d.err", prior_info->oname, it+1);
		writeErrorSinogram(fname, y, Ax, sino_info);
		sprintf(fname, "%s_%d.nmlerr", prior_info->oname, it+1);
		writeNormalizedErrorSinogram(fname, y, Ax, sino_info);*/

		cost = negLogPosterior2D(x, y, Ax, img_info, sino_info, prior_info);
		avg_update = total_updates/(ENTRY)N;
		fprintf(stdout, "cost = %f, avg_update = %f HU\n", cost, mu2hu(avg_update, ATT_AIR, img_info->mu_water));

/*		sprintf(fname, "%s_cost.txt", prior_info->oname);
		writeCost(fname, it, cost, avg_update, img_info->mu_water);*/

		if (avg_update < stop_threshold)
		{
			stop_FLAG = 1;
		}
	}

	sprintf(fname, "%s.rec", prior_info->oname);
	writeImage(fname, image);
/*	sprintf(fname, "%s.err", prior_info->oname);
	writeErrorSinogram(fname, y, Ax, sino_info);
	sprintf(fname, "%s.nmlerr", prior_info->oname);
	writeNormalizedErrorSinogram(fname, y, Ax, sino_info);*/

	if (stop_FLAG == 1)
	{
		fprintf(stdout, "Reached stopping condition.\n");
	}
	else if (stop_threshold > 0)
	{
		fprintf(stdout, "WARNING: Didn't reach stopping condition.\n");
		fprintf(stdout, "Average update magnitude = %f\n", avg_update);
	}
	else
	{
		fprintf(stdout, "No stopping condition.\n");
		fprintf(stdout, "Average update magnitude = %f\n", avg_update);
	}

	free((void *)Ax);
	free((void *)A_col.val);
	free((void *)A_col.index);
	free_img((void **)profiles);
	free((void *)order);
}
